import sys
import requests
import json
import time
import math
import os, os.path
from pony import orm
import collections

from config import *
from model import *
from log import *

if not os.path.exists(JSON_CACHE_DIR):
    os.makedirs(JSON_CACHE_DIR)

def pretty_json(j):
    return json.dumps(j,indent=3,sort_keys=True)

def weight_movie(movie):
    vote = float(movie.get("vote_average",0.0)) + 1.0 
    count = float(movie.get("vote_count",0.0)) + 1.0
    pop = float(movie.get("popularity",0.0)) + 1.0
    return vote * math.log(count) * math.log(pop)

def queue_object(object_id, object_type, object_name='', is_referenced=False, weight=None):
    with orm.db_session:
        if not ObjectRetrieval.get(object_id=object_id, object_type=object_type):
            o = ObjectRetrieval(object_id=object_id, 
                                object_type=object_type,
                                object_name=object_name,
                                is_referenced=is_referenced,
                                weight=weight,
                                is_retrieved=False)
            fine(' ---- queuing', o)

def find_genre(g):
    if type(g) == int:
        genre_id = g
    elif type(g) == dict:
        genre_id = g['id']
    else:
        error('unrecognized object type', g, type(g))
        return Null
    with orm.db_session:
        return Genre[genre_id]

# def find_movie(movie):
#     if type(movie) == int:
#         movie_id = movie
#         movie_name = ''
#     elif type(movie) == dict:
#         movie_id = movie['id']
#         movie_name = movie['title']
#     else:
#         error('unrecognized object type', movie, type(movie))
#         return Null
#     with orm.db_session:
#         if not Movie.get(id=movie_id):
#             Movie(id=movie_id)
#             queue_object(movie_id,'MOVIE',movie_name)
            
#         return Movie[movie_id]

def find_director(director):
    if type(director) == int:
        director_id = director
        director_name = ''
    elif type(director) == dict:
        director_id = director['id']
        director_name = director['name']
    else:
        error('unrecognized object type', director, type(director))
        return Null
    with orm.db_session:
        if not Director.get(id=director_id):
            Director(id=director_id)
            queue_object(director_id,'DIRECTOR',director_name,True)
            
        return Director[director_id]



def find_actor(actor):
    if type(actor) == int:
        actor_id = actor
        actor_name = ''
    elif type(actor) == dict:
        actor_id = actor['id']
        actor_name = actor['name']
    else:
        error('unrecognized object type', actor, type(actor))
        return Null
    with orm.db_session:
        if not Actor.get(id=actor_id):
            Actor(id=actor_id)
            queue_object(actor_id,'ACTOR',actor_name,True)
            
        return Actor[actor_id]



retry_after = None

def tmdb_query(url, **args):
    global retry_after
    retry_after = None
    params = dict([('api_key',API_KEY)] + list(args.items()))
    while (True):
        response =  requests.get(url, params = params)
        if response.status_code == 200:
            return response.json()
        elif response.status_code == 404:
            return collections.defaultdict(lambda : None)
        elif response.status_code == 429:
            try:
                retry_after = int(response.headers['Retry-After']) + 1
            except:
                retry_after = 10
            message('TMDB: Too many requests')
        else:
            warning('TMDB: Unknown response status code returned',response.status_code,response)
            retry_after = 10
        message('TMDB: SLEEPING for',retry_after,'sec')
        time.sleep(retry_after)

def tmdb_get_genres():
    message('TMDB: Retrieving genre list')
    return tmdb_query(BASE_URL + '/genre/movie/list')['genres']

def tmdb_search_movies(query,page=1):
    message('TMDB: Performing search',query,"page",page)
    return tmdb_query(BASE_URL+'/search/movie', query=query,page=page)['results']

def tmdb_get_movie(movie_id):
    if os.path.exists(JSON_CACHE_DIR+('/movie-%d.json'%(movie_id,))):
        #message('CACHE: Retrieving movie details',movie_id)
        return json.load(open(JSON_CACHE_DIR+('/movie-%d.json'%(movie_id,)),'r'))
    else:
        message('TMDB: Retrieving movie details',movie_id)
        movie = tmdb_query(BASE_URL+'/movie/' + str(movie_id), append_to_response='credits')
        open(JSON_CACHE_DIR+('/movie-%d.json'%(movie_id,)),'w').write(pretty_json(movie))
        return movie

def tmdb_get_person(person_id):
    if os.path.exists(JSON_CACHE_DIR+('/person-%d.json'%(person_id,))):
        # message('CACHE: Retrieving person info',person_id)
        return json.load(open(JSON_CACHE_DIR+('/person-%d.json'%(person_id,))))
    else:
        message('TMDB: Retrieving person info',person_id)
        person = tmdb_query(BASE_URL + '/person/' + str(person_id),append_to_response='movie_credits')
        open(JSON_CACHE_DIR+('/person-%d.json'%(person_id,)),'w').write(pretty_json(person))
    return person


# print(pretty_json(tmdb_search_movies('stars',1000)))
# sys.exit(1)

def load_serches():
    message('LOADING SEARCHES')
    with orm.db_session:
        for search_text in open(SEARCH_FILE,'r'):
            if not Search.get(search_text=search_text):
                Search(search_text=search_text)

def store_genres(genres):
    message('LOADING GENRES')
    with orm.db_session:
        for genre in genres:
            if not Genre.get(id=genre['id']):
                fine('registering genre', genre)
                Genre(id=genre['id'],name=genre['name'])


def run_movie_searches():
    while (True):
        with orm.db_session:
            if not Search.exists(lambda s: s.performed_on is None):
                return
            s = Search.select(lambda s: s.performed_on is None).first()
        info('SEARCHING',s.search_text)
        page = 1
        while (page <= MAX_SEARCH_PAGE):
            movies = tmdb_search_movies(s.search_text,page)
            if not movies:
                break
            with orm.db_session:
                for movie in movies:
                    queue_object(movie['id'],'MOVIE',movie['title'],False,weight_movie(movie))
            page += 1
        with orm.db_session:
            Search[s.id].performed_on = datetime.datetime.now() 


def retrieve_object(o):
    if o.object_type == 'MOVIE':
        retrieve_movie(o.object_id)
    elif o.object_type == 'DIRECTOR':
        retrieve_director(o.object_id)
    elif o.object_type == 'ACTOR':
        retrieve_actor(o.object_id)
    else:
        error('Attempting to retrieve unrecognized object type',o.object_type)

def retrieve_movie(movie_id):
    if not Movie.get(id=movie_id):
        Movie(id=movie_id)
    m = Movie[movie_id]
    if m.title is not None:
        warning("Attempting to retrieve an existing movie", m)

    movie = tmdb_get_movie(movie_id)

    m.title        = movie.get('title', '__UNK__')
    m.budget       = movie.get('budget', None)
    m.overview     = movie.get('overview', None)
    m.popularity   = movie.get('popularity', None)
    try: 
        m.release_date = datetime.datetime.strptime(movie['release_date'],'%Y-%m-%d')
    except:
        pass
    m.runtime      = movie.get('runtime', None)
    m.language     = movie.get('original_language', None)
    m.revenue      = movie.get('revenue', None)
    m.vote_average = movie.get('vote_average', None)
    m.vote_count   = movie.get('vote_count', None)

    if 'adult' in movie and type(movie['adult']) == bool:
        m.adult = movie['adult']

    m.genres = [find_genre(g) for g in movie.get('genres',[])]

    for cast in movie.get('credits',{}).get('cast',[]):
        if 'id' in cast:
            Star(movie=m,actor=find_actor(cast),character=cast.get('character',None))
    director_ids = list()
    for crew in movie.get('credits',{}).get('crew',[]):
        if crew.get('job',None) == 'Director' and 'id' in crew:
            director_ids.append(crew['id'])

    m.directors = [find_director(director_id) for director_id in director_ids]

def retrieve_director(director_id):
    if not Director.get(id=director_id):
        Director(id=director_id)
    d = Director[director_id]
    if d.name is not None:
        warning("Attempting to retrieve an existing director", d)

    direct = tmdb_get_person(director_id)
    
    d.name = direct.get('name','__UNK__')
    try:
        d.dob = datetime.datetime.strptime(direct['birthday'],'%Y-%m-%d')
    except:
        pass 
    d.popularity = direct.get('popularity',None)
    d.bio        = direct.get('biography',None)
    d.gender     = direct.get('gender',None)

    if 'adult' in direct and type(direct['adult']) == bool:
        d.adult = direct['adult']

    
    if type(direct.get('movie_credits',None)) == dict and 'crew' in direct['movie_credits']:
        for crew in direct['movie_credits']['crew']:
            queue_object(crew['id'],'MOVIE',crew['title'],False,weight_movie(crew))

def retrieve_actor(actor_id):
    if not Actor.get(id=actor_id):
        Actor(id=actor_id)
    a = Actor[actor_id]
    if a.name is not None:
        warning("Attempting to retrieve an existing actor", a)

    act = tmdb_get_person(actor_id)
    
    a.name = act.get('name','__UNK__')
    try:
        a.dob = datetime.datetime.strptime(act['birthday'],'%Y-%m-%d')
    except:
        pass
    a.popularity = act.get('popularity',None)
    a.bio = act.get('biography',None)
    a.gender = act.get('gender',None)

    if 'adult' in act and type(act['adult']) == bool:
        a.adult = act['adult']

    
    if 'movie_credits' in act and type(act['movie_credits']) == dict and 'cast' in act['movie_credits']:
        for cast in act['movie_credits']['cast']:
            queue_object(cast['id'],'MOVIE',cast['title'],False,weight_movie(cast))


def retrieval_loop():
    stats = count_stats()
    while (True): 
        with orm.db_session:
            movie_count = orm.count(m for m in Movie if m.title is not None)
            ref_count = orm.count(o for o in ObjectRetrieval if o.is_referenced and o.is_retrieved == False)
            queued_movies = orm.count(o for o in ObjectRetrieval if o.object_type == 'MOVIE' and o.is_retrieved == False)
            print_stats(stats)
            if ((movie_count < MOVIE_COUNT_GOAL and queued_movies > 0) or ref_count > 0):
                if ref_count > 0:
                    q = orm.select(o for o in ObjectRetrieval if o.is_retrieved == False and o.is_referenced)
                    o = q.first()
                else:
                    q = orm.select((o,o.weight) for o in ObjectRetrieval if o.is_retrieved == False and o.object_type == 'MOVIE').sort_by(-2)
                    o = q.first()[0]
                    stats = count_stats()

                if o.object_type == 'MOVIE':
                    message(" "*80+"\r %s\n"%str(o),newline=False)
                print_stats(stats)

                retrieve_object(o)
                o.is_retrieved = True
            else:
                break

def count_stats():
    with orm.db_session:
        return dict(movie_count = orm.count(m for m in Movie if m.title is not None),
                    actor_count = orm.count(a for a in Actor if a.name is not None),
                    director_count = orm.count(d for d in Director if d.name is not None),
                    star_count = orm.count(s for s in Star),
                    queued_objects = orm.count(o for o in ObjectRetrieval if o.is_retrieved == False))

def print_stats(stats):
    # with orm.db_session:
    #     movie_count = orm.count(m for m in Movie if m.title is not None)
    #     actor_count = orm.count(a for a in Actor if a.name is not None)
    #     director_count = orm.count(d for d in Director if d.name is not None)
    #     star_count = orm.count(s for s in Star)
    #     queued_objects = orm.count(o for o in ObjectRetrieval if o.is_retrieved == False)
        message("\t%d Queued || %d Mov | %d Act | %d Dir | %d Star \r"%(
            stats['queued_objects'], stats['movie_count'], stats['actor_count'], stats['director_count'], stats['star_count'], ),newline=False)

def display_stats():
    print_stats(count_stats())

if __name__ == '__main__':
    load_serches()
    store_genres(tmdb_get_genres())
    run_movie_searches()
    retrieval_loop()
    display_stats()
